package xmldb;

/**
 * <p>This interface defines the architecture of the
 * index implementations. All implementations have to uniquely index
 * the tokens, i.e. no duplicate tokens are allowed. The following methods
 * are available:</p>
 * <ul>
 * <li>{@link #index(String)} stores new tokens
 * and returns the reference to the indexed instance</li>
 * <li>{@link #get(String)} returns the indexed instance of
 * a token or <code>null</code> if the specified token was not found</li>
 * <li>{@link #size()} returns the number of indexed tokens</li>
 * </ul>
 * <p>The three following methods allow iterating through all the tokens:</p>
 * <ul>
 * <li>{@link #init()} initializes the token iterator</li>
 * <li>{@link #more()} checks if the iterator returns more tokens</li>
 * <li>{@link #next()} returns the next token</li>
 * </ul>
 *
 * <p>XML Technologies WS 15/16 - Assignments<br/>
 * Workgroup DBIS, University of Konstanz</p>
 */
public interface Index {
  /**
   * Checks if the specified token is already indexed; if yes, returns its
   * reference. Otherwise, creates a new index entry and returns a
   * reference of the indexed token.
   * @param token token to be indexed
   * @return indexed token reference
   */
  String index(String token);

  /**
   * Finds the specified token in the index and returns its reference.
   * @param token token to be found
   * @return token or null if token was not found
   */
  String get(String token);

  /**
   * Returns the number of index entries.
   * @return number of index entries
   */
  int size();

  /**
   * Initializes the index iterator.
   */
  void init();

  /**
   * Checks if the iterator has more tokens.
   * @return true if more tokens exist
   */
  boolean more();

  /**
   * Returns the next iterator token.
   * @return next iterator token
   */
  String next();
}
