package xmldb;

import java.io.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * <p>This class scans an XML document and returns its atomic tokens.</p>
 *
 * <p>XML Technologies WS 15/16 - Assignments<br/>
 * Workgroup DBIS, University of Konstanz</p>
 */
public class Scanner extends Constants {
  /** Verbose mode (used for testing). */
  private static final boolean VERBOSE = false;

  /** Scanner States. */
  enum State { CONTENT, TAG, QUOTED, DBL_QUOTED };

  /** Buffered input stream. */
  private BufferedReader input;

  /** Current token type. */
  private Token token;
  /** Current token string. */
  private StringBuilder content;
  /** Current scanner state. */
  private State state = State.CONTENT;

  /** Currently scanned line. */
  private int line = 1;
  /** Currently scanned column. */
  private int col = 1;
  /** Temporary line marker. */
  private int tmpLine;
  /** Temporary column marker. */
  private int tmpCol;
  /** Finished flag. */
  private boolean finished;
  
  /**
   * Main method.
   * @param args command line arguments
   * @throws IOException I/O exception
   */
  public static void main(String[] args) throws IOException {
    // get filename or use default
    String filename = args.length != 0 ? args[0] : INPUT;

    System.out.println("Scanning " + filename);

    // scan document
    Scanner scanner = new Scanner(filename);
    scanner.initTimer();
    while(scanner.next());
    scanner.finish();
    
    scanner.printTimer();
    scanner.printMem();
  }

  /**
   * Constructor; initializes the scanner.
   * @param filename the input file
   * @throws IOException I/O exception
   */
  public Scanner(String filename) throws IOException {
    // retrieves the encoding
    String encoding = getEncoding(filename);
    
    // creates an input stream
    FileInputStream fis = new FileInputStream(filename);
    InputStreamReader isr = new InputStreamReader(fis, encoding);
    input = new BufferedReader(isr);
    
    // skip document declaration and text before first tag (quick+dirty)
    input.mark(2);
    if(input.read() == '<' && input.read() == '?') {
      do input.mark(1); while(input.read() != '<');
    }
    input.reset();
  }
  
  /**
   * Retrieves the encoding from the specified file.
   * @param filename the input file
   * @return encoding
   * @throws IOException I/O exception
   */
  private String getEncoding(String filename) throws IOException {
    FileInputStream in = new FileInputStream(filename);
    String encoding = "UTF-8";
    int b1 = in.read();
    int b2 = in.read();
    
    if(b1 == 0 && b2 == '<') {
      encoding = "UTF-16BE";
    } else if(b1 == '<') {
      if(b2 == 0) {
        encoding = "UTF-16LE";
      } else if(b2 == '?') {
        StringBuilder sb = new StringBuilder();
        while((b1 = in.read()) != '?') sb.append((char) b1);
        String pattern = ".* encoding=[\"'](.*?)[\"'].*";
        Matcher m = Pattern.compile(pattern).matcher(sb.toString());         
        if(m.matches()) encoding = m.group(1);
      }
    }
    in.close();
    return encoding;
  }

  /**
   * Reads and interprets one token from the input stream.
   * @return true if more tokens are found
   * @throws IOException I/O exception
   */
  public boolean next() throws IOException {
    // initialize token content and get next character
    content = new StringBuilder();
    char ch = nextChar();
    if(ch == 0) return false;

    StringBuilder sb;
    if(VERBOSE) {
      // print current scanner state
      sb = new StringBuilder(state.toString());
      while(sb.length() < 14) sb.append(' ');
    }

    // check scanner state
    switch(state) {
      case CONTENT:
        scanCONTENT(ch);
        break;
      case TAG:
        scanTAG(ch);
        break;
      case QUOTED:
        scanATT_VAL(ch);
        break;
      case DBL_QUOTED:
        scanDBL_ATT_VAL(ch);
        break;
    }

    if(VERBOSE) {
      // print found token and string (scanner state can be changed already)
      sb.append(token);
      while(sb.length() < 28) sb.append(' ');
      verbose(sb.toString() + content);
    }
    
    return true;
  }

  /**
   * Finishes scanning.
   * @throws IOException I/O exception
   */
  public void finish() throws IOException {
    input.close();
  }
  
  /**
   * Scans XML content.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanCONTENT(char ch) throws IOException {
    // parse L_BR
    if(ch == '<') scanL_BR(ch);
    // parse TEXT
    else scanTEXT(ch);
  }

  /**
   * Scans XML text.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanTEXT(char ch) throws IOException {
    token = Token.TEXT;
    do {
      content.append(ch);
      ch = nextChar();
    } while(ch != 0 && ch != '<');

    prevChar();
  }

  /**
   * Scans an opening tag.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanL_BR(char ch) throws IOException {
    // start tag was found...
    state = State.TAG;
    token = Token.L_BR;
    content.append(ch);

    // next char...
    ch = nextChar();

    if(ch == '/') {
      // closing tag...
      token = Token.L_BR_CLOSE;
      content.append(ch);
    } else {
      // opening tag...
      prevChar();
    }
  }

  /**
   * Scans a tag content.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanTAG(char ch) throws IOException {
    // scan tag end...
    if(ch == '>') scanR_BR();
    // scan equal sign...
    else if(ch == '=') scanEQ();

    // scan empty tag end...
    else if(ch == '/') scanCLOSE_R_BR(ch);
    // scan quote...
    else if(ch == '\'') scanQUOTE();
    // scan double quote...
    else if(ch == '"') scanDBL_QUOTE();
    // scan whitespace...
    else if(isWhitespace(ch)) scanSPACE(ch);
    // scan tag name...
    else if(isFirstLetter(ch)) scanNAME(ch);
    // undefined character...
    else scanINVALID(ch);
  }

  /**
   * Scans whitespaces.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanSPACE(char ch) throws IOException {
    token = Token.SPACE;
    do {
      content.append(ch);
      ch = nextChar();
    } while(isWhitespace(ch));
    
    prevChar();
  }

  /**
   * Scans the equal sign.
   */
  private void scanEQ() {
    token = Token.EQ;
    content.append('=');
  }

  /**
   * Scans opening quotes.
   */
  private void scanQUOTE() {
    token = Token.QUOTE;
    content.append('\'');
    state = State.QUOTED;
  }

  /**
   * Scan opening double quotes.
   */
  private void scanDBL_QUOTE() {
    token = Token.QUOTE;
    content.append('"');
    state = State.DBL_QUOTED;
  }

  /**
   * Scans closing quotes.
   */
  private void scanQUOTE_CLOSE() {
    token = Token.QUOTE;
    content.append('\'');
    state = State.TAG;
  }

  /**
   * Scans closing double quotes.
   */
  private void scanDBLQUOTE_CLOSE() {
    token = Token.QUOTE;
    content.append('\"');
    state = State.TAG;
  }

  /**
   * Scans a closing tag.
   */
  private void scanR_BR() {
    token = Token.R_BR;
    content.append('>');
    state = State.CONTENT;
  }

  /**
   * Scans the end of a tag.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanCLOSE_R_BR(char ch) throws IOException {
    content.append(ch);
    token = Token.CLOSE_R_BR;
    ch = nextChar();

    if(ch != '>') {
      scanINVALID(ch);
      prevChar();
    } else {
      content.append(ch);
      state = State.CONTENT;
    }
  }

  /**
   * Scans a tag or attribute name.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanNAME(char ch) throws IOException {
    token = Token.NAME;
    do {
      content.append(ch);
      ch = nextChar();
    } while(isLetter(ch));
    prevChar();
  }

  /**
   * Scans an attribute value.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanATT_VAL(char ch) throws IOException {
    // check if quote is finished
    if(ch == '\'') {
      scanQUOTE_CLOSE();
    } else {
      token = Token.ATT_VAL;
      do {
        content.append(ch);
        ch = nextChar();
      } while(ch != 0 && ch != '\'');

      prevChar();
    }
  }

  /**
   * Scans a double quoted attribute value.
   * @param ch current character
   * @throws IOException I/O exception
   */
  private void scanDBL_ATT_VAL(char ch) throws IOException {
    // check if quote is finished
    if(ch == '"') {
      scanDBLQUOTE_CLOSE();
    } else {
      token = Token.ATT_VAL;
      do {
        content.append(ch);
        ch = nextChar();
      } while(ch != 0 && ch != '"');

      prevChar();
    }
  }

  /**
   * Prints a warning for an invalid token.
   * @param ch current character
   */
  private void scanINVALID(char ch) {
    token = Token.INVALID;
    content.append(ch);
    verbose("Invalid Token: line " + line + ", col " + col + " (" + ch + ")");
  }

  /**
   * Returns the next character from the input stream.
   * @return next character or 0 for end of file
   * @throws IOException I/O exception
   */
  private char nextChar() throws IOException {
    // remember stream position and current line/column
    input.mark(1);
    tmpLine = line;
    tmpCol = col;

    int i = input.read();
    // stream finished... return 0
    if(i < 0) {
      finished = true;
      return 0;
    }

    // refresh line/column variables
    if(i == '\n') {
      line++;
      col = 1;
    } else {
      col++;
    }
    // return character
    return (char) i;
  }

  /**
   * Jumps one character back.
   * @throws IOException I/O exception
   */
  private void prevChar() throws IOException {
    // jump back one character and reset line/column pointer
    input.reset();
    line = tmpLine;
    col = tmpCol;
  }

  /**
   * Checks if the specified character is an XML 'first letter'.
   * @param ch current character
   * @return result of check
   */
  private boolean isFirstLetter(char ch) {
    return ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z' || ch >= '_';
  }

  /**
   * Checks if the specified character is an XML 'letter'.
   * @param ch current character
   * @return result of check
   */
  private boolean isLetter(char ch) {
    return isFirstLetter(ch) || ch >= '0' && ch <= '9' ||
      ch == '-' || ch == ':' || ch == '.';
  }

  /**
   * Checks if the specified character is an XML 'whitespace'.
   * @param ch current character
   * @return result of check
   */
  private boolean isWhitespace(char ch) {
    return ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n';
  }

  /**
   * Returns the current token.
   * @return current token
   */
  public String getContent() {
    return content.toString();
  }

  /**
   * Returns the token type.
   * @return token type
   */
  public Token getToken() {
    return token;
  }

  /**
   * Returns the current line.
   * @return current line
   */
  public int getLine() {
    return line;
  }

  /**
   * Returns the current column.
   * @return current column
   */
  public int getColumn() {
    return col;
  }

  /**
   * Returns the finished flag.
   * @return finished flag
   */
  public boolean finished() {
    return finished;
  }
}
