package xmldb; 

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Stack;

import xmldb.Attributes;
import xmldb.Node;

/**
 * <p>This is the main class of the project. It constructs ands
 * organizes the data of an XML file in a table structure.</p>
 *
 * <p>XML Technologies WS 15/16 - Assignments<br/>
 * Workgroup DBIS, University of Konstanz</p>
 */
public class XMLDB extends Constants implements ParserEvents {
  /** Verbose mode (used for testing). */
  private static final boolean VERBOSE = false;

  /** The stack organizes unclosed nodes. */
  Stack<Integer> stack;
  /** Table reference. */
  ArrayList<Node> table;

  /** Element tokens (tag names). */
  Index elements;
  /** Content tokens. */
  Index contents;
  /** Attribute tokens. */
  Index attNames;
  /** Attribute value tokens. */
  Index attValues;

  /**
   * Main method.
   * @param args command line arguments.
   * you can specify the input file as argument
   * @throws IOException I/O exception
   */
  public static void main(String[] args) throws IOException {
    // get filename or use default
    String filename = args.length != 0 ? args[0] : INPUT;
    // check if file exists
    if(!new File(filename).exists()) {
      System.out.println("Sorry, the input file \"" + filename +
          "\" could not be found");
      System.out.println("Usage: java XMLDB <input.xml>");
      return;
    }

    System.out.println("Creating database for " + filename);

    // build table
    XMLDB db = new XMLDB();
    db.initTimer();
    db.create(filename);
    db.printTimer();
    db.printMem();

    // write xml table to disk
    db.writeTable();
    System.out.println("Table was saved in \"" + filename + ".table\"");

    // write xml document to disk
    db.writeDocument();
    System.out.println("Document was saved in \"" + filename + ".out\"");
  }

  /**
   * Parses the document and builds the table.
   * @param filename name of the input file
   * @throws IOException I/O exception
   */
  public void create(String filename) throws IOException {
    // initialize data structures
    elements   = newIndex();
    contents   = newIndex();
    attNames   = newIndex();
    attValues  = newIndex();
    stack = new Stack<Integer>();
    table = new ArrayList<Node>();

    // instantiation of the XMLparser
    Parser parser = new Parser(filename, this);
    // add document node as first table entry
    String token = elements.index(new String(filename));
    Node node = new Node(token, Kind.DOC, -1, null);
    // push root node to stack
    stack.add(table.size());
    table.add(node);

    // parse the xml file
    parser.parse();

    // finish table's document entry
    stack.pop();

    // check if stack was emptied
    if(!stack.empty()) error("Tags have not been completely closed.");
  }

  /**
   * Create a new index from the index structure which is
   * specified in index variable.
   * @return xml token index
   */
  public static Index newIndex() {
    try {
      // instantiate index structure
      return INDEX.newInstance();
    } catch(Exception e) {
      e.printStackTrace();
      return null;
    }
  }

  /**
   * Called when the parser finds an element.
   * @param element tag name
   * @param atts element attributes
   */
  public void startElement(String element, Attributes atts) {
    if(VERBOSE) verbose("startElement  " + element +
        (atts != null ? " (w/attributes)" : ""));

    /* 
     * TODO: Exercise 2, index the tag and create a new table node from the
     * specified tag name and attributes.
     */
  }

  /**
   * Called when an element is finished.
   * @param element tag name
   */
  public void endElement(String element) {
    if(VERBOSE) verbose("endElement    " + element);

    /* 
     * TODO: Exercise 2, complete element node and check if
     * opening and closing tags are equal.
     */
  }

  /**
   * Called when a content is found.
   * @param content string content
   */
  public void characters(String content) {
    if(VERBOSE) verbose("characters    " + content);

    // minimize content size and skip storage of empty contents
    content = content.trim();
    if(content.length() == 0) return;

    /* 
     * TODO: Exercise 2, index and add content (text node) to the table
     */
  }

  /**
   * Writes the table to disk.
   * @throws IOException I/O exception
   */
  public void writeTable() throws IOException {
    // get filename from xml table
    String filename = table.get(0).content + ".table";

    // create writer
    FileWriter fw = new FileWriter(filename);
    BufferedWriter writer = new BufferedWriter(fw);

    // write table header
    writer.write("TYPE      PRE     PAR  CONTENT\n");

    // loop through all table entries
    int size = table.size();
    for(int pre = 0; pre < size; pre++) {
      Node node = table.get(pre);
      formatTable(writer, node.kind.toString(), 5, true);
      formatTable(writer, Integer.toString(pre), 8, false);
      formatTable(writer, Integer.toString(node.parent), 8, false);
      writer.write("  ");
      writer.write(node.content);
      // check attributes
      if(node.atts != null) {
        writer.write(" ");
        for(int i = 0; i < node.atts.size(); i++) {
          String[] att = node.atts.get(i);
          writer.write(att[0] + "=\"" + att[1] + "\" ");
        }
      }
      writer.write('\n');
    }
    writer.close();
  }

  /**
   * Formats the output string.
   * @param writer buffered writer
   * @param text text to be formatted
   * @param size size of table entry
   * @param left alignment (left/right)
   * @throws IOException I/O exception
   */
  private void formatTable(Writer writer, String text, int size, boolean left)
      throws IOException {

    if(left) writer.write(text);
    while(size-- > text.length()) writer.write(' ');
    if(!left) writer.write(text);
  }

  /**
   * Writes the document to disk.
   * @throws IOException I/O exception
   */
  public void writeDocument() throws IOException {
    // get xml document name from table
    String filename = table.get(0).content + ".out";

    // specify output stream
    FileWriter fw = new FileWriter(filename);
    BufferedWriter writer = new BufferedWriter(fw);
    writeDoc(writer);
    writer.close();
  }

  /**
   * Writes the document to the specified writer.
   * @param writer writer reference
   * @throws IOException I/O exception
   */
  public void writeDoc(Writer writer) throws IOException {
	  /*
	   * TODO: Exercise 3, Loop through all table entries and create an XML
	   * output from the table contents:
	   * 
	   * - before printing a node, check if opened tags have to be closed first
	   * - close all opened tags at the end of the table
	   * - get sure that the output will be well-formed
	   * - add some indentation to subordinate nodes, you can use the indent(...)
	   *   function
	   * - print empty tags in their compact representation (<BR/>)
	   */
  }

  /**
   * Adds spaces for xml output.
   * @param writer writer reference
   * @param level level depth
   * @throws IOException I/O exception
   */
  private void indent(Writer writer, int level) throws IOException {
    writer.write('\n');
    for(int j = 0; j < level * 2; j++) writer.write(' ');
  }
}
